/*
 *  CDOF 1.1 Copyright (c) 2004 Cort
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#include <FeatureMgr.h>
#include <FloatMgr.h>
#include "cdof.h"

/* Structure for Lens database */
struct LensStruct {
  char Name[20];
  double FocalLength;
  double CoC;
};

/* Structure for Preferences */
struct PrefStruct {
  UInt16 LensSelect;
  UInt16 Aperture;
  UInt16 FocusType;
  UInt32 Distance;
  UInt32 Distance1;
  UInt32 Distance2;
  UInt32 Distance3;
  UInt32 Distance4;
  UInt32 Distance5;
  UInt32 Distance6;
};

struct PermPrefStruct {
  UInt16 CalOnDrag;
  Boolean EnglishUnit;
};

/* Conversion factor */

#define ft2m 0.3048
#define in2m 0.0254

/* Various constants */

const char *DistPBStrSi1 = ".5m";
const char *DistPBStrSi2 = "1m";
const char *DistPBStrSi3 = "3m";
const char *DistPBStrSi4 = "10m";
const char *DistPBStrSi5 = "30m";
const char *DistPBStrSi6 = "90m";

const char *DistPBStrE1 = "1'";
const char *DistPBStrE2 = "3'";
const char *DistPBStrE3 = "9'";
const char *DistPBStrE4 = "30'";
const char *DistPBStrE5 = "90'";
const char *DistPBStrE6 = "300'";

const UInt16 DistPBMaxSi1 = 500;
const UInt16 DistPBMaxSi2 = 1000;
const UInt16 DistPBMaxSi3 = 3000;
const UInt16 DistPBMaxSi4 = 10000;
const UInt16 DistPBMaxSi5 = 30000;
const UInt16 DistPBMaxSi6 = 9000;

const UInt16 DistPBMaxE1 = 305;
const UInt16 DistPBMaxE2 = 915;
const UInt16 DistPBMaxE3 = 2744;
const UInt16 DistPBMaxE4 = 9144;
const UInt16 DistPBMaxE5 = 27432;
const UInt16 DistPBMaxE6 = 9144;

/* Global variables (preferences) */

static UInt16 gRecordIndex;
static double ApValue_A[] = {22,  19,  16,  13,  11,  9.5,
                             8,   6.7, 5.6, 4.8, 4,   3.3,
                             2.8, 2.4, 2,   1.7, 1.4};
static UInt16 ApPB_A[] = {ApPB22, ApPB19, ApPB16, ApPB13, ApPB11, ApPB95,
                          ApPB8,  ApPB67, ApPB56, ApPB48, ApPB4,  ApPB33,
                          ApPB28, ApPB24, ApPB2,  ApPB17, ApPB14};
#define ApA_Count (sizeof ApValue_A / sizeof ApValue_A[0])
char FocusNearPtr[MAXLEN], FocusPtr[MAXLEN], FocusFarPtr[MAXLEN];

char thouSep, decSep;

static struct PrefStruct Pref;
static struct PermPrefStruct PermPref;

LocalID LensID;
DmOpenRef LensDB;

/****************************************************************************
    The beginning of the PrintDouble section.
 ****************************************************************************/

/*
 *  The following code is written by Fred Bayer, the author of LispMe and
 *  modified by Warren Young, the author of f/Calc.
 *
 *  The following section is released as Public Domain. More information
 *  about the original code can be found here:
 *  http://www.lispme.de/lispme/gcc/tech.html
 */

#define _DONT_USE_FP_TRAPS_ 1
#include <PalmTypes.h>
#include <PalmCompatibility.h>
#include <CoreTraps.h>
#include <StringMgr.h>

void SysTrapFlpLToF(FlpDouble *, Long) SYS_TRAP(sysTrapFlpEmDispatch);

/* convert a long to double */
static double longToDouble(long l) {
  FlpCompDouble fcd;
  asm("moveq.l %0,%%d2": : "i"(sysFloatEm_d_itod):"d2");
  SysTrapFlpLToF(&fcd.fd, l);
  return fcd.d;
}

Long SysTrapFlpFToL(FlpDouble) SYS_TRAP(sysTrapFlpEmDispatch);

/* convert a double to long */
static long doubleToLong(double d) {
  FlpCompDouble fcd;
  fcd.d = d;
  asm("moveq.l %0,%%d2": : "i"(sysFloatEm_d_dtoi):"d2");
  return SysTrapFlpFToL(fcd.fd);
}

void SysTrapBinOp(FlpDouble *, FlpDouble, FlpDouble) SYS_TRAP(sysTrapFlpEmDispatch);

/* the same interface is used for all basic arithmetic operations */
static double genericDoubleOp(double a, double b, long opcode) {
  FlpCompDouble fcda, fcdb, fcds;
  fcda.d = a;
  fcdb.d = b;
  asm("move.l %0,%%d2": : "g"(opcode):"d2");
  SysTrapBinOp(&fcds.fd, fcda.fd, fcdb.fd);
  return fcds.d;
}

/* basic arithmetic operations */
#define addDouble(a,b) genericDoubleOp(a,b,sysFloatEm_d_add)
#define subDouble(a,b) genericDoubleOp(a,b,sysFloatEm_d_sub)
#define mulDouble(a,b) genericDoubleOp(a,b,sysFloatEm_d_mul)
#define divDouble(a,b) genericDoubleOp(a,b,sysFloatEm_d_div)

SDWord SysTrapCompare(FlpDouble, FlpDouble) SYS_TRAP(sysTrapFlpEmDispatch);

/* compare 2 doubles for equality */
Boolean eqDouble(double a, double b) {
  FlpCompDouble fcda, fcdb;
  fcda.d = a;
  fcdb.d = b;
  asm("moveq.l %0,%%d2": : "i"(sysFloatEm_d_feq):"d2");
  return SysTrapCompare(fcda.fd, fcdb.fd);
}

/* compare 2 doubles for less or equal */
static Boolean leqDouble(double a, double b) {
  FlpCompDouble fcda, fcdb;
  fcda.d = a;
  fcdb.d = b;
  asm("moveq.l %0,%%d2": : "i"(sysFloatEm_d_fle):"d2");
  return SysTrapCompare(fcda.fd, fcdb.fd);
}


/**********************************************************************/
/* Formatting parameters                                              */
/**********************************************************************/
#define NUM_DIGITS   15
#define MIN_FLOAT    4
#define ROUND_FACTOR 1.0000000000000005 /* NUM_DIGITS zeros */


/**********************************************************************/
/* FP conversion constants                                            */
/**********************************************************************/
static double pow1[] = {
  1e256, 1e128, 1e064,
  1e032, 1e016, 1e008,
  1e004, 1e002, 1e001
};

static double pow2[] = {
  1e-256, 1e-128, 1e-064,
  1e-032, 1e-016, 1e-008,
  1e-004, 1e-002, 1e-001
};

static void printDouble(double x, char *s) {
  FlpCompDouble fcd;
  short e, e1, i;
  double *pd, *pd1;
  char sign = '\0';
  short dec = 0;

  /*------------------------------------------------------------------*/
  /* Round to desired precision                                       */
  /* (this doesn't always provide a correct last digit!)              */
  /*------------------------------------------------------------------*/
  x = mulDouble(x, ROUND_FACTOR);

  /*------------------------------------------------------------------*/
  /* check for NAN, +INF, -INF, 0                                     */
  /*------------------------------------------------------------------*/
  fcd.d = x;
  if ((fcd.ul[0] & 0x7ff00000) == 0x7ff00000)
    if (fcd.fdb.manH == 0 && fcd.fdb.manL == 0)
      if (fcd.fdb.sign)
        StrCopy(s, "[-inf]");
      else
        StrCopy(s, "[inf]");
    else
      StrCopy(s, "[nan]");
  else if (FlpIsZero(fcd))
    StrCopy(s, "0");
  else {
    /*----------------------------------------------------------------*/
    /* Make positive and store sign                                   */
    /*----------------------------------------------------------------*/
    if (FlpGetSign(fcd)) {
      *s++ = '-';
      FlpSetPositive(fcd);
    }

    if ((unsigned) fcd.fdb.exp < 0x3ff) {   /* meaning x < 1.0 */
      /*--------------------------------------------------------------*/
      /* Build negative exponent                                      */
      /*--------------------------------------------------------------*/
      for (e = 1, e1 = 256, pd = pow1, pd1 = pow2; e1;
           e1 >>= 1, ++pd, ++pd1)
        if (!leqDouble(*pd1, fcd.d)) {
          e += e1;
          fcd.d = mulDouble(fcd.d, *pd);
        }
      fcd.d = mulDouble(fcd.d, 10.0);

      /*--------------------------------------------------------------*/
      /* Only print big exponents                                     */
      /*--------------------------------------------------------------*/
      if (e <= MIN_FLOAT) {
        *s++ = '0';
        *s++ = '.';
        dec = -1;
        while (--e)
          *s++ = '0';
      }
      else
        sign = '-';
    }
    else {
      /*--------------------------------------------------------------*/
      /* Build positive exponent                                      */
      /*--------------------------------------------------------------*/
      for (e = 0, e1 = 256, pd = pow1, pd1 = pow2; e1;
           e1 >>= 1, ++pd, ++pd1)
        if (leqDouble(*pd, fcd.d)) {
          e += e1;
          fcd.d = mulDouble(fcd.d, *pd1);
        }
      if (e < NUM_DIGITS)
        dec = e;
      else
        sign = '+';
    }

    /*----------------------------------------------------------------*/
    /* Extract decimal digits of mantissa                             */
    /*----------------------------------------------------------------*/
    for (i = 0; i < NUM_DIGITS; ++i, --dec) {
      Long d = doubleToLong(fcd.d);
      *s++ = d + '0';
      if (!dec)
        *s++ = '.';
      fcd.d = subDouble(fcd.d, longToDouble(d));
      fcd.d = mulDouble(fcd.d, 10.0);
    }

    /*----------------------------------------------------------------*/
    /* Remove trailing zeros and decimal point                        */
    /*----------------------------------------------------------------*/
    while (s[-1] == '0')
      *--s = '\0';
    if (s[-1] == '.')
      *--s = '\0';

    /*----------------------------------------------------------------*/
    /* Append exponent                                                */
    /*----------------------------------------------------------------*/
    if (sign) {
      *s++ = 'e';
      *s++ = sign;
      StrIToA(s, e);
    }
    else
      *s = '\0';
  }
}

/****************************************************************************
    The end of the PrintDouble section.
 ****************************************************************************/


/*
 *  Calculates a hyper-focal distance. This value is used to
 *  calculate the near and far edges.
 *
 *  focalLength - Focal Length in millimeters (for example 35.0).
 *  aperture    - Aperture (for example 22.0).
 *  coc         - CoC value. This is 0.033 for 35mm film.
 *                0.06 = 6x6cm film
 *                0.15 = 4x5in film
 */
static double CalcHyperfocalDist(double focalLength,
                                 double aperture,
                                 double coc) {
  double retval;
  double fl;

  fl = focalLength;

  retval = (fl * fl) / (aperture * coc);

  return retval;
}

/*
 *  Calculates the near edge.
 *
 *  hyperfocalDist - Use the result of CalcHyperfocalDist().
 *  distOfTarget   - Distance of the actual target.
 *  focalLength    - Focal Length (same as in CalcHyperfocalDist()).
 */
static double CalcNearEdge(double hyperfocalDist,
                           double distOfTarget,
                           double focalLength) {
  double retval;

  retval = (hyperfocalDist * distOfTarget) / (hyperfocalDist +
           (distOfTarget - focalLength));

  return retval;
}


/*
 *  Calculates the far edge.
 *
 *  hyperfocalDist - Use the result of CalcHyperfocalDist().
 *  distOfTarget   - Distance of the actual target.
 *  focalLength    - Focal Length (same as in CalcHyperfocalDist()).
 */
static double CalcFarEdge(double hyperfocalDist,
                          double distOfTarget,
                          double focalLength) {
  double retval;

  retval = (hyperfocalDist * distOfTarget) / (hyperfocalDist -
           (distOfTarget - focalLength));

  return retval;
}

/*
 *  Calculates the Distance using the near edge.
 *
 *  hyperfocalDist - Use the result of CalcHyperfocalDist().
 *  nearedge       - Distance of the near edge.
 *  focalLength    - Focal Length (same as in CalcHyperfocalDist()).
 */
static double CalcDistFNear(double hyperfocalDist,
                           double nearedge,
                           double focalLength) {
  double retval;

  retval = (nearedge * (focalLength - hyperfocalDist)) /
           (nearedge - hyperfocalDist);

  return retval;
}

/*
 *  Calculates the Distance using the far edge.
 *
 *  hyperfocalDist - Use the result of CalcHyperfocalDist().
 *  faredge        - Distance of the far edge.
 *  focalLength    - Focal Length (same as in CalcHyperfocalDist()).
 */
static double CalcDistFFar(double hyperfocalDist,
                           double faredge,
                           double focalLength) {
  double retval;

  retval = (faredge * (focalLength + hyperfocalDist)) /
           (faredge + hyperfocalDist);

  return retval;
}

/*
 *  This will cut the decimals. No roundings.
 */

static double CutDecimals(double d) {
  long l;

  l = (long) (d * 100.0);
  return (double) l / 100.0;
}

/*
 *  Returns pointer to the object by pointer ID from the given form.
 */
static void* GetObjectPtr(FormPtr formP, UInt16 objectID)
{
  FormPtr frmP;
  
  if (formP != NULL) {
    frmP = formP;
  }
  else {
    frmP = FrmGetActiveForm();
  }
  
  return FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID));  
}

/*
 *  Display the result
 */
void DisplayResult(void *ObjPtr, MemPtr tempPtr, FlpCompDouble input) {
  char temp[MAXLEN];
  FlpCompDouble value;
  double limit1,limit2;
  
  if (PermPref.EnglishUnit) {
    limit1 = 0.3048;
    limit2 = 16093.44;
  }
  else {
    limit1 = 1.0;
    limit2 = 10000.0;
  }

  if (leqDouble(input.d,0.0) && !eqDouble(input.d,0.0))
    StrCopy(tempPtr, "INF");
  else if (leqDouble(input.d,limit1)) {
    if (PermPref.EnglishUnit)
      value.d = CutDecimals(input.d / in2m);
    else
      value.d = CutDecimals(input.d * 100.0);

    printDouble(value.d, temp);
    StrLocalizeNumber(temp, thouSep, decSep);
    StrCopy(tempPtr, temp);

    if (PermPref.EnglishUnit)
      StrCat(tempPtr, "\"");
    else
      StrCat(tempPtr, "cm");
  }
  else if (leqDouble(input.d,limit2)) {
    if (PermPref.EnglishUnit)
      value.d = CutDecimals(input.d / ft2m);
    else
      value.d = CutDecimals(input.d);

    printDouble(value.d, temp);
    StrLocalizeNumber(temp, thouSep, decSep);
    StrCopy(tempPtr, temp);

    if (PermPref.EnglishUnit)
      StrCat(tempPtr, "'");
    else
      StrCat(tempPtr, "m");
  }
  else
    if (PermPref.EnglishUnit)
      StrCopy(tempPtr, ">10miles");
    else
      StrCopy(tempPtr, ">10km");

  CtlSetLabel(ObjPtr, tempPtr);
}

/*
 *  The calculate result.
 */
void CalcResult(FormPtr frmP) {
  double hyp;
  FlpCompDouble near, far, focus;
  double FocalLength = 0.0;
  double CoC = 0.0;
  struct LensStruct *Lens;
  MemHandle LensMH;
  
  if (Pref.LensSelect != noListSelection) {
    LensMH = DmQueryRecord(LensDB, Pref.LensSelect);
    Lens = (struct LensStruct*) MemHandleLock(LensMH);
    FocalLength = Lens->FocalLength;
    CoC = Lens->CoC;
    MemHandleUnlock(LensMH);
  }

  if (Pref.Distance == 0 || eqDouble(FocalLength,0.0) || eqDouble(CoC,0.0)) {
    focus.d = 0;
    near.d = 0;
    far.d = 0;
  }
  else {
    hyp = CalcHyperfocalDist(FocalLength, ApValue_A[Pref.Aperture], CoC);
    switch (Pref.FocusType) {
    case 0:
      focus.d = CalcDistFNear(hyp, Pref.Distance, FocalLength) / 1000.0;
      near.d = Pref.Distance / 1000.0;
      far.d = CalcFarEdge(hyp, focus.d * 1000.0, FocalLength) / 1000.0;
      break;
    case 1:
      focus.d = Pref.Distance / 1000.0;
      near.d = CalcNearEdge(hyp, Pref.Distance, FocalLength) / 1000.0;
      far.d = CalcFarEdge(hyp, Pref.Distance, FocalLength) / 1000.0;
      break;
    case 2:
      focus.d = CalcDistFFar(hyp, Pref.Distance, FocalLength) / 1000.0;
      near.d = CalcNearEdge(hyp, focus.d * 1000.0, FocalLength) / 1000.0;
      far.d = Pref.Distance / 1000.0;
      break;
    }
  }

  DisplayResult(GetObjectPtr(frmP,FocusNearPB), FocusNearPtr, near);
  DisplayResult(GetObjectPtr(frmP,FocusPB), FocusPtr, focus);
  DisplayResult(GetObjectPtr(frmP,FocusFarPB), FocusFarPtr, far);
}

/*
 *  The main form event handler.
 */
static Boolean AppFormHandleEventMain(EventPtr event) {
  FormPtr frmP;
  Boolean handled = true;
  static UInt16 FirstRepeat = 1;
  UInt32 RecordCount, a, b;
  FlpCompDouble tempdouble;
  MemHandle tempMH;
  struct LensStruct *Lens;
  static char **LensText = NULL;
  
  frmP = FrmGetActiveForm();

  /* Keys */
  if (event->eType == keyDownEvent) {
      handled = false;
  }

  /* Controls */
  else if (event->eType == ctlSelectEvent) {
    /* Slider */
    switch (event->data.ctlSelect.controlID) {
    case DistSlide50: case DistSlide100: case DistSlide300:
    case DistSlide1000: case DistSlide3000: case DistSlide9000:
      FirstRepeat = 1;
      CtlSetValue(GetObjectPtr(frmP,DistPB50),0);
      CtlSetValue(GetObjectPtr(frmP,DistPB100),0);
      CtlSetValue(GetObjectPtr(frmP,DistPB300),0);
      CtlSetValue(GetObjectPtr(frmP,DistPB1000),0);
      CtlSetValue(GetObjectPtr(frmP,DistPB3000),0);
      CtlSetValue(GetObjectPtr(frmP,DistPB9000),0);

      if (event->data.ctlSelect.controlID == DistSlide9000)
        Pref.Distance = (UInt32) event->data.ctlSelect.value * 10;
      else
        Pref.Distance = event->data.ctlSelect.value;

      switch (event->data.ctlSelect.controlID) {
      case DistSlide50:
        CtlSetValue(GetObjectPtr(frmP,DistPB50),1);
        Pref.Distance1 = Pref.Distance;
        break;
      case DistSlide100:
        CtlSetValue(GetObjectPtr(frmP,DistPB100),1);
        Pref.Distance2 = Pref.Distance;
        break;
      case DistSlide300:
        CtlSetValue(GetObjectPtr(frmP,DistPB300),1);
        Pref.Distance3 = Pref.Distance;
        break;
      case DistSlide1000:
        CtlSetValue(GetObjectPtr(frmP,DistPB1000),1);
        Pref.Distance4 = Pref.Distance;
        break;
      case DistSlide3000:
        CtlSetValue(GetObjectPtr(frmP,DistPB3000),1);
        Pref.Distance5 = Pref.Distance;
        break;
      case DistSlide9000:
        CtlSetValue(GetObjectPtr(frmP,DistPB9000),1);
        Pref.Distance6 = Pref.Distance;
        break;
      }

      handled = true;
    }

    switch (event->data.ctlSelect.controlID) {
      case DistPB50:
        Pref.Distance = CtlGetValue(GetObjectPtr(frmP,DistSlide50));
        break;
      case DistPB100:
        Pref.Distance = CtlGetValue(GetObjectPtr(frmP,DistSlide100));
        break;
      case DistPB300:
        Pref.Distance = CtlGetValue(GetObjectPtr(frmP,DistSlide300));
        break;
      case DistPB1000:
        Pref.Distance = CtlGetValue(GetObjectPtr(frmP,DistSlide1000));
        break;
      case DistPB3000:
        Pref.Distance = CtlGetValue(GetObjectPtr(frmP,DistSlide3000));
        break;
      case DistPB9000:
        Pref.Distance = (UInt32) CtlGetValue(GetObjectPtr(frmP,DistSlide9000)) * 10;
        break;

      case FocusNearPB:
        Pref.FocusType = 0;
        break;
      case FocusPB:
        Pref.FocusType = 1;
        break;
      case FocusFarPB:
        Pref.FocusType = 2;
        break;

      default:
        handled = false;
        break;
    }

    for (a=0; a<ApA_Count; a++)
      if (event->data.ctlSelect.controlID == ApPB_A[a]) {
        Pref.Aperture=a;
        handled = true;
        break;
      }

    CalcResult(frmP);
  }

  /* Repeat event for slider */
  else if (event->eType == ctlRepeatEvent) {
    switch (event->data.ctlRepeat.controlID) {
    case DistSlide50: case DistSlide100: case DistSlide300:
    case DistSlide1000: case DistSlide3000: case DistSlide9000:

      if (FirstRepeat) {
        FirstRepeat = 0;
        CtlSetValue(GetObjectPtr(frmP,DistPB50),0);
        CtlSetValue(GetObjectPtr(frmP,DistPB100),0);
        CtlSetValue(GetObjectPtr(frmP,DistPB300),0);
        CtlSetValue(GetObjectPtr(frmP,DistPB1000),0);
        CtlSetValue(GetObjectPtr(frmP,DistPB3000),0);
        CtlSetValue(GetObjectPtr(frmP,DistPB9000),0);

        switch (event->data.ctlRepeat.controlID) {
        case DistSlide50:
          CtlSetValue(GetObjectPtr(frmP,DistPB50),1);
          break;
        case DistSlide100:
          CtlSetValue(GetObjectPtr(frmP,DistPB100),1);
          break;
        case DistSlide300:
          CtlSetValue(GetObjectPtr(frmP,DistPB300),1);
          break;
        case DistSlide1000:
          CtlSetValue(GetObjectPtr(frmP,DistPB1000),1);
          break;
        case DistSlide3000:
          CtlSetValue(GetObjectPtr(frmP,DistPB3000),1);
          break;
        case DistSlide9000:
          CtlSetValue(GetObjectPtr(frmP,DistPB9000),1);
          break;
        }
      }

      if (event->data.ctlRepeat.controlID == DistSlide9000)
        Pref.Distance = (UInt32) event->data.ctlRepeat.value * 10;
      else
        Pref.Distance = event->data.ctlRepeat.value;

      handled = false;
      if (PermPref.CalOnDrag)
        CalcResult(frmP);
      else {
        tempdouble.d = Pref.Distance / 1000.0;
        switch (Pref.FocusType) {
        case 0:
          DisplayResult(GetObjectPtr(frmP,FocusNearPB), FocusNearPtr, tempdouble);
          break;
        case 1:
          DisplayResult(GetObjectPtr(frmP,FocusPB), FocusPtr, tempdouble);
          break;
        case 2:
          DisplayResult(GetObjectPtr(frmP,FocusFarPB), FocusFarPtr, tempdouble);
          break;
        }
      }
    }
  }

  /* Menus */
  else if (event->eType == menuEvent) {
    switch (event->data.menu.itemID) {
      case MenuItemPref:
        FrmGotoForm(PrefForm);
        handled = true;
        break;
      case MenuItemAbout:
        FrmAlert(AlertAbout);
        handled = true;
        break;
      case MenuItemHow:
        FrmHelp(3);
        handled = true;
        break;
      default:
        handled = false;
    }
  }

  /* Popup lists */
  else if (event->eType == popSelectEvent) {
    switch (event->data.popSelect.listID) {
    case LensL:
      a = event->data.popSelect.selection;
      if (a == LstGetNumberOfItems(event->data.popSelect.listP) - 1) {
        FrmGotoForm(LensListForm);
        handled = true;
      }
      else {
        Pref.LensSelect = event->data.popSelect.selection;
        CalcResult(frmP);
        handled = false;
      }
      break;
    default:
      handled = false;
      break;
    }
  }

  /* Form open */
  else if (event->eType == frmOpenEvent) {
    /* Initialize Lens Popup List */
    DmDatabaseSize(0,LensID,&RecordCount,NULL,NULL);

      /* Allocate memory to store string */
    LensText = MemPtrNew(sizeof(*LensText) * (RecordCount + 1));
      
    for (a=0; a<(RecordCount+1); a++) {
      LensText[a] = (char*) MemPtrNew(sizeof(char) * 20);
    }
    
    for (a=0; a<RecordCount; a++) {
      tempMH=DmQueryRecord(LensDB,a);
      Lens = (struct LensStruct*) MemHandleLock(tempMH);
      StrCopy(LensText[a], Lens->Name);
      MemHandleUnlock(tempMH);
    }
    
    StrCopy(LensText[RecordCount], "Edit Lens...");
    
    LstSetListChoices(GetObjectPtr(frmP,LensL),LensText,RecordCount+1);
    LstSetHeight(GetObjectPtr(frmP,LensL),RecordCount+1);
    if (Pref.LensSelect >= RecordCount)
      Pref.LensSelect = noListSelection;
    LstSetSelection(GetObjectPtr(frmP,LensL),Pref.LensSelect);
    if (Pref.LensSelect >= 0 && Pref.LensSelect < RecordCount)
      CtlSetLabel(GetObjectPtr(frmP,LensPU),LensText[Pref.LensSelect]);

    /* Initialize Aperture */
    for (a=0; a<ApA_Count; a++)
      CtlSetValue(GetObjectPtr(frmP,ApPB_A[a]),0);

    CtlSetValue(GetObjectPtr(frmP,ApPB_A[Pref.Aperture]),1);

    /* Initialize Distance */
    
    if (PermPref.EnglishUnit) {
      CtlSetLabel(GetObjectPtr(frmP,DistPB50), DistPBStrE1);
      CtlSetLabel(GetObjectPtr(frmP,DistPB100), DistPBStrE2);
      CtlSetLabel(GetObjectPtr(frmP,DistPB300), DistPBStrE3);
      CtlSetLabel(GetObjectPtr(frmP,DistPB1000), DistPBStrE4);
      CtlSetLabel(GetObjectPtr(frmP,DistPB3000), DistPBStrE5);
      CtlSetLabel(GetObjectPtr(frmP,DistPB9000), DistPBStrE6);

      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide50),NULL,&DistPBMaxE1,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide100),NULL,&DistPBMaxE2,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide300),NULL,&DistPBMaxE3,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide1000),NULL,&DistPBMaxE4,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide3000),NULL,&DistPBMaxE5,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide9000),NULL,&DistPBMaxE6,NULL,NULL);

      if (Pref.Distance1 > DistPBMaxE1)
        Pref.Distance1 = DistPBMaxE1;
      if (Pref.Distance2 > DistPBMaxE2)
        Pref.Distance2 = DistPBMaxE2;
      if (Pref.Distance3 > DistPBMaxE3)
        Pref.Distance3 = DistPBMaxE3;
      if (Pref.Distance4 > DistPBMaxE4)
        Pref.Distance4 = DistPBMaxE4;
      if (Pref.Distance5 > DistPBMaxE5)
        Pref.Distance5 = DistPBMaxE5;
      if (Pref.Distance6 > DistPBMaxE6)
        Pref.Distance6 = DistPBMaxE6;
    }
    else {
      CtlSetLabel(GetObjectPtr(frmP,DistPB50), DistPBStrSi1);
      CtlSetLabel(GetObjectPtr(frmP,DistPB100), DistPBStrSi2);
      CtlSetLabel(GetObjectPtr(frmP,DistPB300), DistPBStrSi3);
      CtlSetLabel(GetObjectPtr(frmP,DistPB1000), DistPBStrSi4);
      CtlSetLabel(GetObjectPtr(frmP,DistPB3000), DistPBStrSi5);
      CtlSetLabel(GetObjectPtr(frmP,DistPB9000), DistPBStrSi6);

      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide50),NULL,&DistPBMaxSi1,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide100),NULL,&DistPBMaxSi2,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide300),NULL,&DistPBMaxSi3,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide1000),NULL,&DistPBMaxSi4,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide3000),NULL,&DistPBMaxSi5,NULL,NULL);
      CtlSetSliderValues(GetObjectPtr(frmP,DistSlide9000),NULL,&DistPBMaxSi6,NULL,NULL);

      if (Pref.Distance1 > DistPBMaxSi1)
        Pref.Distance1 = DistPBMaxSi1;
      if (Pref.Distance2 > DistPBMaxSi2)
        Pref.Distance2 = DistPBMaxSi2;
      if (Pref.Distance3 > DistPBMaxSi3)
        Pref.Distance3 = DistPBMaxSi3;
      if (Pref.Distance4 > DistPBMaxSi4)
        Pref.Distance4 = DistPBMaxSi4;
      if (Pref.Distance5 > DistPBMaxSi5)
        Pref.Distance5 = DistPBMaxSi5;
      if (Pref.Distance6 > DistPBMaxSi6)
        Pref.Distance6 = DistPBMaxSi6;
    }

    CtlSetValue(GetObjectPtr(frmP,DistSlide50),Pref.Distance1);
    CtlSetValue(GetObjectPtr(frmP,DistSlide100),Pref.Distance2);
    CtlSetValue(GetObjectPtr(frmP,DistSlide300),Pref.Distance3);
    CtlSetValue(GetObjectPtr(frmP,DistSlide1000),Pref.Distance4);
    CtlSetValue(GetObjectPtr(frmP,DistSlide3000),Pref.Distance5);
    CtlSetValue(GetObjectPtr(frmP,DistSlide9000),Pref.Distance6 / 10);

    if (Pref.Distance == Pref.Distance1)
      CtlSetValue(GetObjectPtr(frmP,DistPB50),1);
    else if (Pref.Distance == Pref.Distance2)
      CtlSetValue(GetObjectPtr(frmP,DistPB100),1);
    else if (Pref.Distance == Pref.Distance3)
      CtlSetValue(GetObjectPtr(frmP,DistPB300),1);
    else if (Pref.Distance == Pref.Distance4)
      CtlSetValue(GetObjectPtr(frmP,DistPB1000),1);
    else if (Pref.Distance == Pref.Distance5)
      CtlSetValue(GetObjectPtr(frmP,DistPB3000),1);
    else
      CtlSetValue(GetObjectPtr(frmP,DistPB9000),1);

    /* Initialize Focus Type */
    CtlSetValue(GetObjectPtr(frmP,FocusNearPB),0);
    CtlSetValue(GetObjectPtr(frmP,FocusPB),0);
    CtlSetValue(GetObjectPtr(frmP,FocusFarPB),0);
    switch (Pref.FocusType) {
    case 0:
      CtlSetValue(GetObjectPtr(frmP,FocusNearPB),1);
      break;
    case 1:
      CtlSetValue(GetObjectPtr(frmP,FocusPB),1);
      break;
    case 2:
      CtlSetValue(GetObjectPtr(frmP,FocusFarPB),1);
      break;
    }
    
    /* Finally, calculate the result with what we have */
    CalcResult(frmP);

    FrmDrawForm(frmP);
    handled = true;
  }

  /* Form close */
  else if (event->eType == frmCloseEvent) {

    /* Free previously allocated memory if any */
    if (LensText != NULL) {
      b = MemPtrSize(LensText) / sizeof(*LensText);
      for (a=0; a<b; a++)
        MemPtrFree(LensText[a]);
      MemPtrFree(LensText);
    }
    LensText = NULL;

    handled = false;
  }
  else {
    handled = false;
  }
  
  return handled;
}

static Boolean AppFormHandleEventLensList(EventPtr event) {
  FormPtr frmP;
  Boolean handled = true;
  static UInt16 LensSelect;
  UInt16 tempSelect, InitializeList = 0;
  UInt32 RecordCount, a, b;
  MemHandle tempMH;
  struct LensStruct *Lens;
  static char **LensText = NULL;
  
  frmP = FrmGetActiveForm();

  /* Controls */
  if (event->eType == ctlSelectEvent) {
    switch (event->data.ctlSelect.controlID) {
      case LensLNewB:
        gRecordIndex = dmMaxRecordIndex;
        FrmGotoForm(LensEditForm);
        handled = true;
        break;
      case LensLCancelB:
        FrmGotoForm(MainForm);
        handled = true;
        break;
      case LensLUpB:
        tempSelect = LstGetSelection(GetObjectPtr(frmP,LensLL));
        if (tempSelect != noListSelection) {
          if (tempSelect != 0) {
            DmMoveRecord(LensDB, tempSelect, tempSelect-1);
            LstSetSelection(GetObjectPtr(frmP,LensLL), tempSelect-1);
            LensSelect = tempSelect - 1;
          }
        }
        handled = true;
        InitializeList = 1;
        break;
      case LensLDownB:
        tempSelect = LstGetSelection(GetObjectPtr(frmP,LensLL));
        if (tempSelect != noListSelection) {
          if (tempSelect != LstGetNumberOfItems(GetObjectPtr(frmP,LensLL))-1) {
            DmMoveRecord(LensDB, tempSelect, tempSelect+2);
            LstSetSelection(GetObjectPtr(frmP,LensLL), tempSelect+1);
            LensSelect = tempSelect + 1;
          }
        }
        handled = true;
        InitializeList = 1;
        break;
      default:
        handled = false;
    }
  }

  /* List selection */
  else if (event->eType == lstSelectEvent) {
    switch (event->data.lstSelect.listID) {
      case LensLL:
        if (LensSelect == event->data.lstSelect.selection) {
          if (event->data.lstSelect.selection != noListSelection) {
            gRecordIndex = event->data.lstSelect.selection;
            FrmGotoForm(LensEditForm);
          }
        }
        else
          LensSelect = event->data.lstSelect.selection;
        handled = true;
        break;
      default:
        handled = false;
    }
  }

  /* Form open */
  else if (event->eType == frmOpenEvent) {
    FrmDrawForm(frmP);
    LensSelect = noListSelection;
    handled = true;
    InitializeList = 1;
  }

  /* Form close */
  else if (event->eType == frmCloseEvent) {

    /* Free previously allocated memory if any */
    if (LensText != NULL) {
      b = MemPtrSize(LensText) / sizeof(*LensText);
      for (a=0; a<b; a++)
        MemPtrFree(LensText[a]);
      MemPtrFree(LensText);
    }
    LensText = NULL;

    handled = false;
  }

  else {
    handled = false;
  }

  if (InitializeList) {
    /* Initialize Lens Popup List */
    DmDatabaseSize(0,LensID,&RecordCount,NULL,NULL);

    /* Free previously allocated memory if any */
    if (LensText != NULL) {
      b = MemPtrSize(LensText) / sizeof(*LensText);
      for (a=0; a<b; a++)
        MemPtrFree(LensText[a]);
      MemPtrFree(LensText);
    }
    LensText = NULL;

      /* Allocate memory to store string */
    if (RecordCount != 0) {
      LensText = MemPtrNew(sizeof(*LensText) * RecordCount);
    }

    for (a=0; a<RecordCount; a++) {
      LensText[a] = (char*) MemPtrNew(sizeof(char) * 20);
      tempMH=DmQueryRecord(LensDB,a);
      Lens = (struct LensStruct*) MemHandleLock(tempMH);
      StrCopy(LensText[a], Lens->Name);
      MemHandleUnlock(tempMH);
    }

    LstSetListChoices(GetObjectPtr(frmP,LensLL),LensText,RecordCount);
    LstDrawList(GetObjectPtr(frmP,LensLL));
  }
  
  return handled;
}

static Boolean AppFormHandleEventLensEdit(EventPtr event) {
  FormPtr frmP;
  Boolean handled = true;
  struct LensStruct Lens, *LensRec;
  char *tempstr;
  FlpCompDouble value;
  MemHandle LensMH, tempMH;
  
  frmP = FrmGetActiveForm();

  /* Controls */
  if (event->eType == ctlSelectEvent) {
    switch (event->data.ctlSelect.controlID) {
      case LensEDoneB:
        tempstr = FldGetTextPtr(GetObjectPtr(frmP,LensENameF));
        if (tempstr == NULL) {
          FrmAlert(AlertNoName);
          break;
        }
        StrCopy(Lens.Name,tempstr);

        tempstr = FldGetTextPtr(GetObjectPtr(frmP,LensEFocalF));
        if (tempstr == NULL) {
          FrmAlert(AlertNoFocal);
          break;
        }
        StrDelocalizeNumber(tempstr, thouSep, decSep);
        FlpBufferAToF(&value.fd, tempstr);
        if (eqDouble(value.d,0.0)) {
          FrmAlert(AlertNoFocal);
          break;
        }
        Lens.FocalLength = value.d;
        
        tempstr = FldGetTextPtr(GetObjectPtr(frmP,LensECoCF));
        if (tempstr == NULL) {
          FrmAlert(AlertNoCoC);
          break;
        }
        StrDelocalizeNumber(tempstr, thouSep, decSep);
        FlpBufferAToF(&value.fd, tempstr);
        if (eqDouble(value.d,0.0)) {
          FrmAlert(AlertNoCoC);
          break;
        }
        Lens.CoC = value.d;

        if (gRecordIndex == dmMaxRecordIndex)
          LensMH = DmNewRecord(LensDB, &gRecordIndex, sizeof(Lens));
        else
          LensMH = DmGetRecord(LensDB, gRecordIndex);
        LensRec = MemHandleLock(LensMH);
        DmWrite(LensRec,0,&Lens,sizeof(Lens));
        MemHandleUnlock(LensMH);
        DmReleaseRecord(LensDB, gRecordIndex, true);

        FrmGotoForm(LensListForm);

        break;        
      case LensECancelB:
        FrmGotoForm(LensListForm);
        handled = true;
        break;
      case LensEDeleteB:
        if (gRecordIndex == dmMaxRecordIndex) {
          FrmGotoForm(LensListForm);
          handled = true;
          break;
        }
        if (FrmAlert(AlertDeleteLens) == 0) {
          DmRemoveRecord(LensDB,gRecordIndex);
          FrmGotoForm(LensListForm);
        }
        handled = true;
        break;
      default:
        handled = false;
    }
  }

  /* Form open */
  else if (event->eType == frmOpenEvent) {
    /* Edit an existing record */
    if (gRecordIndex != dmMaxRecordIndex) {
      LensMH = DmQueryRecord(LensDB, gRecordIndex);
      LensRec = MemHandleLock(LensMH);

      /* Initialize Name */
      tempMH = MemHandleNew(sizeof(char) * 20);
      tempstr = MemHandleLock(tempMH);
      StrCopy(tempstr,LensRec->Name);
      MemHandleUnlock(tempMH);
      FldSetTextHandle(GetObjectPtr(frmP,LensENameF),tempMH);
      
      /* Initialize FocalLength */
      tempMH = MemHandleNew(sizeof(char) * MAXLEN);
      tempstr = MemHandleLock(tempMH);
      printDouble(LensRec->FocalLength, tempstr);
      StrLocalizeNumber(tempstr, thouSep, decSep);
      MemHandleUnlock(tempMH);
      FldSetTextHandle(GetObjectPtr(frmP,LensEFocalF),tempMH);
        		
      /* Initialize CoC */
      tempMH = MemHandleNew(sizeof(char) * MAXLEN);
      tempstr = MemHandleLock(tempMH);
      printDouble(LensRec->CoC, tempstr);
      StrLocalizeNumber(tempstr, thouSep, decSep);
      MemHandleUnlock(tempMH);
      FldSetTextHandle(GetObjectPtr(frmP,LensECoCF),tempMH);

      MemHandleUnlock(LensMH);
    }  

    FrmDrawForm(frmP);
    handled = true;
  }

  else {
    handled = false;
  }
  
  return handled;
}

static Boolean AppFormHandleEventPref(EventPtr event) {
  FormPtr frmP;
  Boolean handled = true;
  
  frmP = FrmGetActiveForm();

  /* Controls */
  if (event->eType == ctlSelectEvent) {
    switch (event->data.ctlSelect.controlID) {
      case PrefDoneB:
        PermPref.CalOnDrag = CtlGetValue(GetObjectPtr(frmP,PrefCalOnDragCB));
        PermPref.EnglishUnit = CtlGetValue(GetObjectPtr(frmP,PrefEnglishUnitCB));
        FrmGotoForm(MainForm);
        handled = true;
        break;
      case PrefCancelB:
        FrmGotoForm(MainForm);
        handled = true;
        break;
      default:
        handled = false;
    }
  }

  /* Form open */
  else if (event->eType == frmOpenEvent) {
    CtlSetValue(GetObjectPtr(frmP,PrefCalOnDragCB),PermPref.CalOnDrag);
    CtlSetValue(GetObjectPtr(frmP,PrefEnglishUnitCB),PermPref.EnglishUnit);
    FrmDrawForm(frmP);
    handled = true;
  }

  else {
    handled = false;
  }
  
  return handled;
}

/*
 *  Application event handler.
 */
static Boolean AppHandleEvent(EventPtr event) {
  Boolean eventHandled = true;
  FormPtr frm;
  
  if (event->eType == frmLoadEvent) {
    frm = FrmInitForm(event->data.frmLoad.formID);
    FrmSetActiveForm(frm);
    
    switch (event->data.frmLoad.formID) {
    case MainForm:
      FrmSetEventHandler(frm, AppFormHandleEventMain);
      break;
    case LensListForm:
      FrmSetEventHandler(frm, AppFormHandleEventLensList);
      break;
    case LensEditForm:
      FrmSetEventHandler(frm, AppFormHandleEventLensEdit);
      break;
    case PrefForm:
      FrmSetEventHandler(frm, AppFormHandleEventPref);
      break;
    default:
      eventHandled = false;
      break;
    }
  }
  else {
    eventHandled = false;
  }
  
  return eventHandled;
}

/*
 *  Application event loop.
 */
static void AppEventLoop() {
  UInt16 error;
  EventType event;
  
  do {
    EvtGetEvent(&event, evtWaitForever);
    if (!SysHandleEvent(&event)) {
      if (!MenuHandleEvent(NULL, &event, &error)) {
        if (!AppHandleEvent(&event)) {
          FrmDispatchEvent(&event);
        }
      }
    }
  } while (event.eType != appStopEvent);
}

/*
 *  Start-up function.
 */
static UInt16 StartApplication() {
  NumberFormatType numFormat;
  UInt16 size;
  UInt16 version;
  Err error;

  size = sizeof(PermPref);
  version = PrefGetAppPreferences(CreatorID, 0, &PermPref, &size, true);

  if (version != 1 || size != sizeof(PermPref)) {
    PermPref.CalOnDrag = 0;
    PermPref.EnglishUnit = 0;
  }
  
  size = sizeof(Pref);
  version = PrefGetAppPreferences(CreatorID, 0, &Pref, &size, false);

  if (version != 1 || size != sizeof(Pref)) {
    Pref.LensSelect = noListSelection;
    Pref.Aperture = 0;
    Pref.FocusType = 1;
    Pref.Distance = 0;
    Pref.Distance1 = 0;
    Pref.Distance2 = 0;
    Pref.Distance3 = 0;
    Pref.Distance4 = 0;
    Pref.Distance5 = 0;
    Pref.Distance6 = 0;
  }

  /* Open the lens definition database */

  LensID = DmFindDatabase(0, "CDoF_LensDB");
  if (LensID == 0)
  {
    error = DmCreateDatabase(0,"CDoF_LensDB",CreatorID,DBType,false);
    if (error != errNone)
      return error;
    LensID = DmFindDatabase(0, "CDoF_LensDB");
    version = 0;
    error = DmSetDatabaseInfo(0, LensID, NULL, NULL, &version, NULL,
                              NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    if (error != errNone)
      return error;
  }
  else {
    error = DmDatabaseInfo(0, LensID, NULL, NULL, &version, NULL,
                           NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    if (error != errNone)
      return error;

    /* FIXME: Need a better way to deal with incorrect db version */
    if (version != 0)
      return 1;
  }

  LensDB = DmOpenDatabase(0, LensID, dmModeReadWrite);
  if (LensDB == 0)
    return DmGetLastErr();

  /* Check the number format */
  numFormat = (NumberFormatType) PrefGetPreference(prefNumberFormat);
  LocGetNumberSeparators(numFormat, &thouSep, &decSep);

  return 0;
}

/*
 *  Shutdown function.
 */
static void StopApplication() {

  PrefSetAppPreferences(CreatorID, 0, 1, &PermPref, sizeof(PermPref), true);
  PrefSetAppPreferences(CreatorID, 0, 1, &Pref, sizeof(Pref), false);
  DmCloseDatabase(LensDB);

  FrmCloseAllForms();
}

/*
 *  The main function.
 */
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags) {
  UInt16 error;
  UInt32 romVersion;

  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);

  if (romVersion < VERSION) {
    return sysErrRomIncompatible;
  }
  
  if (cmd == sysAppLaunchCmdNormalLaunch) {
    error = StartApplication();
    if (error) {
      return error;
    }
    FrmGotoForm(MainForm);
    AppEventLoop();
    StopApplication();
  }
  
  return 0;
}
